/***************************************************************************
Copyright (c) 2016, Guillaume Baud-Berthier

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
***************************************************************************/

#include "bmc.h"

using namespace Glucose;

unsigned UNDEF_LIT = INT_MAX;

float enco_cumul_time = 0;


/* ========================================================================== */
/*                           BMC algorithm functions                          */
/* ========================================================================== */

int bmc(int start, int stop, aiger *model, bool inprep_hl, bool nosimp, bool verbose_) {
  int i, j;
  int status = -1;
  clock_t clk;


  if (verbose_) clk = clock();

  if (verbose_) printf("Mem: %d MB\n", get_mem());

  unsigned bad_lit = init_model(model, stop, 0, verbose_);

  hl_ve(verbose_);

  if (verbose_) {
    clk = clock() - clk;
    enco_cumul_time += (float)(clk)/(float)(CLOCKS_PER_SEC);
  }

  for(i = 0; i <= stop; i++) {

    if (verbose_) printf("Mem: %d MB\n", get_mem());
    if (verbose_) printf("[bmc]  Encoding %d..\n", i);
    if (verbose_) clk = clock();

    if (inprep_hl) {
      int simp_prev = import_unit_from_solver(verbose_);
      int tmp_simp_prev = import_binary_from_solver(verbose_);
      simp_prev = simp_prev < tmp_simp_prev ? simp_prev : tmp_simp_prev;
      if (simp_prev != INT_MAX) {
    	clear_global_hash();
	for (int k = 0; k < simp_prev; k++)
	  fill_global_hash(k);
    	for (int k = simp_prev; k < i; k++)
    	  simp_ands(k, bad_lit, 0, verbose_, 0);
      }
    }

    list_ands.push_back(unroll(i));
    
    iter_flatten_ptu(i);

    if (!nosimp)
      simp_ands(i, bad_lit, 0, verbose_, 0);
    
    ands2clauses_coi(bad_lit, i, verbose_);

    // /* Solver simplification */
    // simp_with_solver(i, mapdef[aiger_lit2var(bad_lit)]->lhs_ts[i] >> 1);

    // /* TEST */
    // bool simp_prev = import_unit_from_solver(verbose_);
    // simp_prev |= import_binary_from_solver(verbose_);
    // if (simp_prev) {
    //   clear_global_hash();
    //   for (int k = 0; k <= i; k++) {
    // 	simp_ands(k, bad_lit, 0, verbose_, elim_ll);
    //   }
    // }

    /********/
    /*************************/


    if (verbose_) {
      clk = clock() - clk;
      enco_cumul_time += (float)(clk)/(float)(CLOCKS_PER_SEC);
    }
    
    if (verbose_) { printf("[bmc]  Solving..\n"); fflush(stdout); }

    if ((!mapdef[aiger_lit2var(bad_lit)]->lhs_ts[i] &&
	 !aiger_sign(bad_lit)) ||
	(mapdef[aiger_lit2var(bad_lit)]->lhs_ts[i] == 1 &&
	 aiger_sign(bad_lit))) {
      
      if (verbose_) printf("[bmc]  UNSAT (Solved by simplification)\n");
      printf("u%d\n", i);
      fflush(stdout);

      if (!dll_size(dll_latches)) {
	status = 0;
	break;
      }

    } else if ((!mapdef[aiger_lit2var(bad_lit)]->lhs_ts[i] &&
		aiger_sign(bad_lit)) ||
	       (mapdef[aiger_lit2var(bad_lit)]->lhs_ts[i] == 1 &&
		!aiger_sign(bad_lit))) {

      if (verbose_) printf("[bmc]  CEX (Solved by simplification)\n");
      status = 1;
      break;
	
    } else {

      if (verbose_) clk = clock();

      unsigned var_po = mapdef[aiger_lit2var(bad_lit)]->lhs_ts[i];
      assert(var_po > 1);
      std::map<unsigned, ands_def*>::iterator it_lit_ref = map_ands.find(aiger_lit2var(var_po));

      assert(it_lit_ref != map_ands.end()); // Might be false if po <=> input
      assert(it_lit_ref->second->cst == -1);
      
      int lit_bad = addPO(bad_lit, i);

      assume(mapdef[0]->lhs_ts.back() >> 1);


      if (verbose_) {
	clk = clock() - clk;
	enco_cumul_time += (float)(clk)/(float)(CLOCKS_PER_SEC);
      }
      
      lbool res_check;
      res_check = check_sat(verbose_);
      if (res_check == l_True) {
	status = 1;
	break;
      } else if (res_check == l_Undef) {
	assert(0);
      }


      printf("u%d\n", i);
      fflush(stdout);

      if (!dll_size(dll_latches)) {
	status = 0;
	break;
      }   
      
      //printf("UCL %d\n", -(mapdef[0]->lhs_ts.back() >> 1));
      //ucl(-(mapdef[0]->lhs_ts.back() >> 1)); //
      //printf("UCL %d\n", -lit_bad);
      ucl(-lit_bad); //

      // TODO: Ajouter la constante -lit_bad dans la représentation du modèle déplié
      // TBD: Se fait automatiquement avec l'importation (A VERIFIER)
    }
  }

  
  printf("%d\nb0\n", status);
  fflush(stdout);
  printf(".\n");

  if (verbose_) {
    printf("Encoding %.2fs | Simplification %.2fs | Solving %.2fs\n", enco_cumul_time,
	   solver->simp_cumul_time, solver->solv_cumul_time);
  }
  fflush(stdout);

  if (verbose_) printf("Mem: %d MB\n", get_mem());

  //destroy_model(verbose_);
  return 0;
}


